# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

class Space(object):

    def __init__(self, shape):
        self.shape = shape
    
    def sample(self):
        raise NotImplementedError


class BuyHoldSellSpace(Space):
    def __init__(self):
        super().__init__(2)
    
    def sample(self,seed = None):
        if seed:
            np.random.seed(seed)
        action = np.zeros(self.shape)
        action[0] = np.random.randint(low=0,high=3)
        action[1] = np.random.randint(low=0,high=100)
        return action


# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# https://becominghuman.ai/genetic-algorithm-for-reinforcement-learning-a38a5612c4dc
# https://gist.github.com/karpathy/77fbb6a8dac5395f1b73e7a89300318d


from collections import deque

import uuid
import ray

from fingym import fingym

import argparse
import numpy as np

import copy
import os

ray.init(num_cpus=12)

NOISE = np.random.randn(250000000).astype(np.float32)

envs = []
spyEnv = fingym.make('SPY-Daily-v0')
envs.append(spyEnv)
# tslaEnv = fingym.make('TSLA-Daily-v0')
# envs.append(tslaEnv)
# googlEnv = fingym.make('GOOGL-Daily-v0')
# envs.append(googlEnv)
# cgcEnv = fingym.make('CGC-Daily-v0')
# envs.append(cgcEnv)
# cronEnv = fingym.make('CRON-Daily-v0')
# envs.append(cronEnv)
# baEnv = fingym.make('BA-Daily-v0')
# envs.append(baEnv)
# amznEnv = fingym.make('AMZN-Daily-v0')
# envs.append(amznEnv)
# amdEnv = fingym.make('AMD-Daily-v0')
# envs.append(amdEnv)
# abbvEnv = fingym.make('ABBV-Daily-v0')
# envs.append(abbvEnv)
# aaplEnv = fingym.make('AAPL-Daily-v0')
#envs.append(aaplEnv)

def build_compile_model(input_size, output_size):
    import tensorflow as tf
    from tensorflow import keras
    from tensorflow.keras import Model, Sequential
    from tensorflow.keras.layers import Dense, Embedding, Reshape
    from tensorflow.keras.optimizers import Adam

    model = Sequential()

    model.add(Dense(400, input_shape=(input_size,), activation='relu',kernel_initializer='he_uniform', use_bias=True, bias_initializer=keras.initializers.Constant(0.1)))
    model.add(Dense(300, activation='relu', kernel_initializer='he_uniform', use_bias=True, bias_initializer=keras.initializers.Constant(0.1)))
    model.add(Dense(output_size, activation='softmax',use_bias=True, bias_initializer=keras.initializers.Constant(0.1)))
    model.compile(loss='categorical_crossentropy', optimizer=Adam(learning_rate=0.01))

    return model

def save_model_weights(model, filepath):
    model.save_weights(filepath)

class EvoAgent():
    def __init__(self, env_state_dim, time_frame, weights, id = None):
        if id is None:
            self.id = uuid.uuid4()
        else:
            self.id = id
        self.state_size = env_state_dim
        self.time_frame = time_frame
        # when this is full, start making predictions
        self.state_fifo = deque(maxlen=self.time_frame)

        # 0 - do nothing
        # 1 - buy w/ multiplier .33
        # 2 - buy w/ multiplier .5
        # 3 - buy w/ multiplier .66
        # 4 - sell w/ multiplier .33
        # 5 - sell w/ multiplier .5
        # 6 - sell w/ multiplier .66
        self.action_size = 7

        self.max_shares_to_trade_at_once = 100

        self.weights = weights
    
    def deep_copy(self, id=None):
        weights = np.copy(self.weights)
        new_agent = EvoAgent(self.state_size, self.time_frame, weights, id)
        return new_agent
    
    def act(self, state, model):
        self.state_fifo.append(state)

        # do nothing for the first time frames until we can start the prediction
        if len(self.state_fifo) < self.time_frame:
            return np.zeros(2)
        
        state = np.array(list(self.state_fifo))
        state = np.reshape(state,(self.state_size*self.time_frame,1))

        output_probabilities = model.predict_on_batch(state.T)[0]
        output_probabilities = np.array(output_probabilities)
        output_probabilities /= output_probabilities.sum()
        try:
            action = np.random.choice(range(self.action_size),1,p=output_probabilities).item()
        except:
            print('output probabilities: ', output_probabilities)
            action = np.zeros(2)
        env_action = self._nn_action_to_env_action(action)
        return env_action
    
    def _nn_action_to_env_action(self,nn_action):
        env_action = [0,0]
        
        if nn_action == 0:
            env_action = [0,0]
        if nn_action == 1:
            env_action = [1, 0.33 * self.max_shares_to_trade_at_once]
        if nn_action == 2:
            env_action = [1, 0.5 * self.max_shares_to_trade_at_once]
        if nn_action == 3:
            env_action = [1, 0.66 * self.max_shares_to_trade_at_once]
        if nn_action == 4:
            env_action = [2, 0.33 * self.max_shares_to_trade_at_once]
        if nn_action == 5:
            env_action = [2, 0.5 * self.max_shares_to_trade_at_once]
        if nn_action == 6:
            env_action = [2, 0.66 * self.max_shares_to_trade_at_once]

        return env_action

def create_random_agents(num_agents, state_size, time_frame):
    agents = []
    for _ in range(num_agents):
        model = build_compile_model(state_size * time_frame, 7)
        agent = EvoAgent(state_size, time_frame, model.get_weights())
        agents.append(agent)

    return agents

def run_agent(env, agent, model):

    state = env.reset()
    # Removed time element from state
    state = np.delete(state, 2)
    state_as_percentages = state
    done = False
    while not done:
        action = agent.act(state_as_percentages, model)
        next_state, reward, done, info = env.step(action)
        if len(next_state) > agent.state_size:
            next_state = np.delete(next_state, 2)
        opn = (next_state[2] - state[2]) / next_state[2]
        high = (next_state[3] - state[3]) / next_state[3]
        low = (next_state[4] - state[4]) / next_state[4]
        close = (next_state[5] - state[5]) / next_state[5]
        volume = (next_state[6] - state[6]) / next_state[6]
        state_as_percentages = [next_state[0], next_state[1], opn, high, low, close, volume]
        state = next_state
    return info['cur_val']

@ray.remote
def return_average_score(envs, agent, runs):
    model = build_compile_model(agent.state_size * agent.time_frame, agent.action_size)
    model.set_weights(agent.weights)
    score = 0
    for env in envs:
        envScore = 0
        for i in range(runs):
            envScore += run_agent(env, agent, model)
        envScore = envScore/runs
        print('AGENT: {}, env: {}, score: {}'.format(agent.id, type(env).__name__,envScore))
        score += envScore
    score = score/len(envs)
    print('AGENT: {}, total score: {}'.format(agent.id,score))
    return agent, score

def run_agents_n_times(envs, agents, runs):
    avg_score_map = {}
    avg_score = []
    futures = [return_average_score.remote(envs,agent,runs) for agent in agents]
    results = ray.get(futures)
    for result in results:
        avg_score_map[result[0].id] = result[1]
    for agent in agents:
        avg_score.append(avg_score_map[agent.id])
    return avg_score

def mutate(agent):
    print('mutate')
    obj_id = ray.put(agent.weights)
    child_agent = EvoAgent(agent.state_size, agent.time_frame, np.copy(ray.get(obj_id)))#ray.get(obj_id).deep_copy()
    
    #child_agent = agent.deep_copy()

    mutation_power = 0.02

    weights = child_agent.weights
    # array copied to object store (objects are immutable), not heap, so a copy must be made
    # https://github.com/ray-project/ray/issues/369
    #weights = np.copy(c_weights)
    #weights.flags.writeable = True
    #print('flags: ', weights.flags)

    for weight in weights:
        
        if len(weight.shape) == 2:
            for i0 in range(weight.shape[0]):
                for i1 in range(weight.shape[1]):
                    #print('flags: ', weight.flags)
                    weight.flags.writeable = True
                    weight[i0,i1]+= mutation_power*NOISE[np.random.randint(0,len(NOISE)-1)]

        if len(weight.shape) == 1:
            for i0 in range(weight.shape[0]):
                weight.flags.writeable = True
                weight[i0]+= mutation_power*NOISE[np.random.randint(0,len(NOISE)-1)]
    
    child_agent.weights = weights
    return child_agent

def add_elite(envs, agents, sorted_parent_indexes, elite_index = None, only_consider_top_n=10):
    candidate_elite_index = sorted_parent_indexes[:only_consider_top_n]

    if(elite_index is not None):
        candidate_elite_index = np.append(candidate_elite_index,[elite_index])

    top_score = None
    top_elite_index = None
    elite_agent = None

    futures = [return_average_score.remote(envs, agents[i],runs=5) for i in candidate_elite_index]
    results = ray.get(futures)
    
    dirname = os.path.dirname(__file__)

    i = 0
    
    for result in results:
        agent = result[0]
        score = result[1]

        print("Score for elite agent {} is {} ".format(agent.id,score))
        model = build_compile_model(agent.state_size * agent.time_frame, agent.action_size)
        model.set_weights(agent.weights)
        save_model_weights(model,os.path.join(dirname,'evo_weights_' + str(i) + '.h5'))

        if(top_score is None):
            top_score = score
            elite_agent = agent
        elif(score > top_score):
            top_score = score
            elite_agent = agent
        
        i+=1
            
    print("Elite selected has id ",elite_agent.id, " and score", top_score)
    
    child_agent = elite_agent.deep_copy(elite_agent.id)
    return child_agent

def return_children(envs, agents, sorted_parent_indexes, elite_index):
    children_agents = []

    for i in range(len(agents) -1):
        selected_agent_index = sorted_parent_indexes[np.random.randint(len(sorted_parent_indexes))]
        children_agents.append(mutate(agents[selected_agent_index]))
        
    
    # now add one elite
    elite_child = add_elite(envs, agents, sorted_parent_indexes, elite_index)
    children_agents.append(elite_child)
    elite_index=len(children_agents)-1

    return children_agents, elite_index

if __name__ == '__main__':

    # removing time element from state_dim
    state_size = envs[0].state_dim - 1
    print('state_size: ', state_size)

    time_frame = 30
    num_agents = 4

    agents = create_random_agents(num_agents, state_size, time_frame)

    # first agent gets saved weights
    dirname = os.path.dirname(__file__)

    # how many top agents to consider as parents
    top_limit = 2

    for i in range (top_limit):
        weights_file = os.path.join(dirname,'evo_weights_' + str(i) + '.h5')
        if os.path.exists(weights_file):
            print('loading existing weights')
            model = build_compile_model(agents[0].state_size * agents[0].time_frame, agents[0].action_size)
            model.load_weights(weights_file)
            agents[i].weights = model.get_weights()

    # run evolution until x generations
    generations = 1000

    elite_index = None

    for generation in range(generations):
        rewards = run_agents_n_times(envs,agents,5) # average of x times

        # sort by rewards
        sorted_parent_indexes = np.argsort(rewards)[::-1][:top_limit]

        top_rewards = []
        for best_parent in sorted_parent_indexes:
            top_rewards.append(rewards[best_parent])
        
        print("Generation ", generation, " | Mean rewards: ", np.mean(rewards), " | Mean of top 5: ",np.mean(top_rewards[:5]))
        print("Top ",top_limit," scores", sorted_parent_indexes)
        print("Rewards for top: ",top_rewards)

        children_agents, elite_index = return_children(envs, agents, sorted_parent_indexes, elite_index)

        agents = children_agents


# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fingym.envs.spy_envs import DailySpyEnv

import os

class TslaDailyEnv(DailySpyEnv):
    def __init__(self):
        super().__init__()
    
    def _get_data_file(self):
        return os.path.join(os.path.dirname(__file__),'..','data/filtered_tsla_data_10_yrs.csv')


# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fingym.gym import make, DailySpyEnv, IntradaySpyEnv, GooglDailyEnv, CronDailyEnv, CgcDailyEnv, BaDailyEnv, AmznDailyEnv, AmdDailyEnv, AbbvDailyEnv, AaplDailyEnv

import pytest

@pytest.fixture
def spy_daily_v0_env():
    return make('SPY-Daily-v0')

@pytest.fixture
def spy_intraday_v0_env():
    return make('SPY-Minute-v0')

@pytest.fixture
def googl_daily_v0_env():
    return make('GOOGL-Daily-v0')

@pytest.fixture
def cron_daily_v0_env():
    return make('CRON-Daily-v0')

@pytest.fixture
def cgc_daily_v0_env():
    return make('CGC-Daily-v0')

@pytest.fixture
def ba_daily_v0_env():
    return make('BA-Daily-v0')

@pytest.fixture
def amzn_daily_v0_env():
    return make('AMZN-Daily-v0')

@pytest.fixture
def amd_daily_v0_env():
    return make('AMD-Daily-v0')

@pytest.fixture
def abbv_daily_v0_env():
    return make('ABBV-Daily-v0')

@pytest.fixture
def aapl_daily_v0_env():
    return make('AAPL-Daily-v0')

def test_make_spy_daily_v0(spy_daily_v0_env):
    assert type(spy_daily_v0_env) == DailySpyEnv

def test_make_spy_intraday_v0(spy_intraday_v0_env):
    assert type(spy_intraday_v0_env) == IntradaySpyEnv

def test_make_googl_daily_v0(googl_daily_v0_env):
    assert type(googl_daily_v0_env) == GooglDailyEnv

def test_make_cron_daily_v0(cron_daily_v0_env):
    assert type(cron_daily_v0_env) == CronDailyEnv

def test_make_cgc_daily_v0(cgc_daily_v0_env):
    assert type(cgc_daily_v0_env) == CgcDailyEnv

def test_make_ba_daily_v0(ba_daily_v0_env):
    assert type(ba_daily_v0_env) == BaDailyEnv

def test_make_amzn_daily_v0(amzn_daily_v0_env):
    assert type(amzn_daily_v0_env) == AmznDailyEnv

def test_make_amd_daily_v0(amd_daily_v0_env):
    assert type(amd_daily_v0_env) == AmdDailyEnv

def test_make_abbv_daily_v0(abbv_daily_v0_env):
    assert type(abbv_daily_v0_env) == AbbvDailyEnv

def test_make_aapl_daily_v0(aapl_daily_v0_env):
    assert type(aapl_daily_v0_env) == AaplDailyEnv


# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# https://becominghuman.ai/genetic-algorithm-for-reinforcement-learning-a38a5612c4dc


from collections import deque

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import model_from_json, clone_model
from tensorflow.keras import Model, Sequential
from tensorflow.keras.layers import Dense, Embedding, Reshape
from tensorflow.keras.optimizers import Adam

from fingym import fingym

import argparse
import numpy as np

import copy
import os

class EvoAgent():
    def __init__(self, env_state_dim, time_frame):
        self.state_size = env_state_dim
        self.time_frame = time_frame
        # when this is full, start making predictions
        self.state_fifo = deque(maxlen=self.time_frame)

        # 0 - do nothing
        # 1 - buy w/ multiplier .33
        # 2 - buy w/ multiplier .5
        # 3 - buy w/ multiplier .66
        # 4 - sell w/ multiplier .33
        # 5 - sell w/ multiplier .5
        # 6 - sell w/ multiplier .66
        self.action_size = 7

        self.max_shares_to_trade_at_once = 100

        self.model = self._build_compile_model()
    
    def deep_copy(self):
        new_agent = EvoAgent(self.state_size, self.time_frame)
        new_agent.model.set_weights(self.model.get_weights())

        return new_agent
    
    def act(self, state):
        self.state_fifo.append(state)

        # do nothing for the first time frames until we can start the prediction
        if len(self.state_fifo) < self.time_frame:
            return np.zeros(2)
        
        state = np.array(list(self.state_fifo))
        state = np.reshape(state,(self.state_size*self.time_frame,1))

        output_probabilities = self.model.predict_on_batch(state.T)[0]
        #output_probabilities /= output_probabilities.sum()
        #print('output_probabilities: ', output_probabilities)
        action = np.random.choice(range(self.action_size),1,p=output_probabilities).item()
        #print('action: ', action)
        env_action = self._nn_action_to_env_action(action)
        return env_action
    
    def save_model_weights(self, filepath):
        self.model.save_weights(filepath)
    
    def _build_compile_model(self):
        model = Sequential()

        input_size = self.state_size * self.time_frame
        model.add(Dense(400, input_shape=(input_size,), activation='relu',kernel_initializer='he_uniform', use_bias=True, bias_initializer=keras.initializers.Constant(0.1)))
        model.add(Dense(300, activation='relu', kernel_initializer='he_uniform', use_bias=True, bias_initializer=keras.initializers.Constant(0.1)))
        #model.add(Dense(1, activation='softmax',use_bias=True))
        model.add(Dense(self.action_size, activation='softmax',use_bias=True, bias_initializer=keras.initializers.Constant(0.1)))
        # we won't really use loss or optimizer for evolutionary agents
        #model.compile(loss='mse', optimizer=Adam(learning_rate=0.01))
        model.compile(loss='categorical_crossentropy', optimizer=Adam(learning_rate=0.01))
        #print(model.summary())
        return model
    
    def _nn_action_to_env_action(self,nn_action):
        env_action = [0,0]
        
        if nn_action == 0:
            env_action = [0,0]
        if nn_action == 1:
            env_action = [1, 0.33 * self.max_shares_to_trade_at_once]
        if nn_action == 2:
            env_action = [1, 0.5 * self.max_shares_to_trade_at_once]
        if nn_action == 3:
            env_action = [1, 0.66 * self.max_shares_to_trade_at_once]
        if nn_action == 4:
            env_action = [2, 0.33 * self.max_shares_to_trade_at_once]
        if nn_action == 5:
            env_action = [2, 0.5 * self.max_shares_to_trade_at_once]
        if nn_action == 6:
            env_action = [2, 0.66 * self.max_shares_to_trade_at_once]

        return env_action

def create_random_agents(num_agents, state_size, time_frame):
    agents = []
    for _ in range(num_agents):
        agent = EvoAgent(state_size, time_frame)
        agents.append(agent)

    return agents

def run_agent(env, agent):
    state = env.reset()
    # Removed time element from state
    state = np.delete(state, 2)
    done = False
    while not done:
        action = agent.act(state)
        next_state, reward, done, info = env.step(action)
        state = next_state
        if len(state) > agent.state_size:
            state = np.delete(state, 2)
    return info['cur_val']

def return_average_score(env, agent, runs):
    score = 0
    print('***** agent score *****')
    for i in range(runs):
        score += run_agent(env, agent)
        print('score: ', score)
    return score/runs

def run_agents_n_times(env, agents, runs):
    avg_score = []
    for agent in agents:
        avg_score.append(return_average_score(env, agent, runs))
    return avg_score

def mutate(agent):
    print('mutate')
    child_agent = agent.deep_copy()

    mutation_power = 0.02

    weights = child_agent.model.get_weights()

    for weight in weights:
        #print('weight len: ', len(weight.shape))
        
        if len(weight.shape) == 2:
            for i0 in range(weight.shape[0]):
                for i1 in range(weight.shape[1]):
                    weight[i0,i1]+= mutation_power*np.random.randn()

        if len(weight.shape) == 1:
            for i0 in range(weight.shape[0]):
                weight[i0]+= mutation_power*np.random.randn()
    
    child_agent.model.set_weights(weights)
    #print('parent_weights: ', agent.model.get_weights())
    #print('child_weights: ', child_agent.model.get_weights())
    return child_agent

def add_elite(env, agents, sorted_parent_indexes, elite_index = None, only_consider_top_n=10):
    candidate_elite_index = sorted_parent_indexes[:only_consider_top_n]

    if(elite_index is not None):
        candidate_elite_index = np.append(candidate_elite_index,[elite_index])

    top_score = None
    top_elite_index = None
    
    for i in candidate_elite_index:
        score = return_average_score(env, agents[i],runs=5)
        print("Score for elite i ", i, " is ", score)
        
        if(top_score is None):
            top_score = score
            top_elite_index = i
        elif(score > top_score):
            top_score = score
            top_elite_index = i
            
    print("Elite selected with index ",top_elite_index, " and score", top_score)
    dirname = os.path.dirname(__file__)
    agents[top_elite_index].save_model_weights(os.path.join(dirname,'evo_weights.h5'))
    
    child_agent = agents[top_elite_index].deep_copy()
    return child_agent

def return_children(env, agents, sorted_parent_indexes, elite_index):
    children_agents = []

    for i in range(len(agents) -1):
        selected_agent_index = sorted_parent_indexes[np.random.randint(len(sorted_parent_indexes))]
        children_agents.append(mutate(agents[selected_agent_index]))
        
    
    # now add one elite
    elite_child = add_elite(env, agents, sorted_parent_indexes, elite_index)
    children_agents.append(elite_child)
    elite_index=len(children_agents)-1

    return children_agents, elite_index

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=None)
    parser.add_argument('env_id', nargs='?', default='SPY-Daily-v0', help='Select the environment to run')
    args = parser.parse_args()

    env = fingym.make(args.env_id)

    # removing time element from state_dim
    state_size = env.state_dim - 1
    print('state_size: ', state_size)

    time_frame = 30
    num_agents = 400

    agents = create_random_agents(num_agents, state_size, time_frame)

    # first agent gets saved weights
    dirname = os.path.dirname(__file__)
    os.path.join(dirname,'evo_weights.h5')
    weights_file=os.path.join(dirname,'evo_weights.h5')
    if os.path.exists(weights_file):
        print('loading existing weights')
        agents[0].model.load_weights(weights_file)

    # how many top agents to consider as parents
    top_limit = 20

    # run evolution until x generations
    generations = 1000

    elite_index = None

    for generation in range(generations):
        rewards = run_agents_n_times(env,agents,5) # average of x times

        # sort by rewards
        sorted_parent_indexes = np.argsort(rewards)[::-1][:top_limit]

        top_rewards = []
        for best_parent in sorted_parent_indexes:
            top_rewards.append(rewards[best_parent])
        
        print("Generation ", generation, " | Mean rewards: ", np.mean(rewards), " | Mean of top 5: ",np.mean(top_rewards[:5]))
        print("Top ",top_limit," scores", sorted_parent_indexes)
        print("Rewards for top: ",top_rewards)

        children_agents, elite_index = return_children(env, agents, sorted_parent_indexes, elite_index)

        agents = children_agents


# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pandas as pd
import numpy as np
import sys
from datetime import datetime
import re

from fingym.envs.env import Env
from fingym.spaces.space import BuyHoldSellSpace

class SpyEnv(Env):
    def __init__(self):

        self.headers, self.data = self._load_data()
        self.n_step = len(self.data)
        self.cur_step = 0
        self.initial_investment = 25000
        self.cash_in_hand = self.initial_investment
        self._close_idx = 4
        self._open_idx = 1
        self.stock_owned = 0

        # stock_owned, cash_in_hand, date, hclo, volume
        self.state_dim = 2 + len(self.headers)

        # action[0]
        # 0 - hold/do nothing
        # 1 - buy
        # 2 - hold
        # action[1]
        # Number of actions to buy/sell
        self.action_size = 2
        self.action_space = BuyHoldSellSpace()
    
    def reset(self):
        self.cur_step = 0
        self.cash_in_hand = self.initial_investment
        self.stock_owned = 0
        return self._get_obs() 

    def step(self, action):
        
        prev_val = self._get_val()
        if self.cur_step < self.n_step -1:
            self.cur_step += 1
        # else:
        #     print('WARNING - this environment is done')
        self._trade(action)

        # get the new value after taking the action
        cur_val = self._get_val()

        reward = cur_val - prev_val
        obs = self._get_obs()
        done = self.cur_step >= self.n_step - 1
        info = {'cur_val': cur_val}


        return obs, reward, done, info

    def _trade(self, action):
        
        action[1] = round(action[1])
        # Mimic buying/selling the next day at open
        # from getting the previous day's prices.
        stock_price = self._get_stock_price_open()

        # sell
        shares_to_sell = 0
        if action[0] == 2:
            shares_to_sell = action[1]
        
        # buy
        shares_to_buy = 0
        if action[0] ==1:
            shares_to_buy = action[1]

        if shares_to_sell > 0:
            # sell all shares
            if self.stock_owned < shares_to_sell:
                self.cash_in_hand += self.stock_owned * stock_price
                self.stock_owned = 0
            else:
                self.stock_owned -= shares_to_sell
                self.cash_in_hand += shares_to_sell * stock_price
        
        
        if shares_to_buy > 0:
            can_buy = True
            shares_bought = 0
            while can_buy:
                if self.cash_in_hand > stock_price and shares_bought < shares_to_buy:
                    self.stock_owned += 1
                    shares_bought += 1
                    self.cash_in_hand -= stock_price
                else:
                    can_buy = False

    
    def _get_obs(self):
        """
        Return array with number stock owned, stock price and cash in hand.
        Ex: [stock_owned, cash_in_hand, date, hclo, volume]
        """
        obs = np.empty(self.state_dim)
        obs[0] = self.stock_owned
        obs[1] = self.cash_in_hand
        # date
        obs[2] = self._convert_date_str_to_datetime(self.data[self.cur_step][0]).timestamp()
        obs[3:] = self.data[self.cur_step][1:]
        return obs

    def _get_val(self):
        return self.stock_owned * self._get_stock_price_close() + self.cash_in_hand
    
    def _get_stock_price_close(self):
        return self.data[self.cur_step][self._close_idx]

    def _get_stock_price_open(self):
        return self.data[self.cur_step][self._open_idx]

    def _load_data(self):
        raise NotImplementedError

    def _convert_date_str_to_datetime(self,datestr):
        raise NotImplementedError

class DailySpyEnv(SpyEnv):
    def __init__(self):
        self.date_str_format = '%m/%d/%Y'
        super().__init__()
    
    def _convert_date_str_to_datetime(self,datestr):
        if sys.version_info[1] <=5:
            datestr = re.sub(r'([-+]\d{2}):(\d{2})(?:(\d{2}))?$', r'\1\2\3', datestr)
        return datetime.strptime(datestr,self.date_str_format)
    
    def _load_data(self):
        data_file = self._get_data_file()
        df = pd.read_csv(data_file)
        return df.columns.values, df.values
    
    def _get_data_file(self):
        return os.path.join(os.path.dirname(__file__),'..','data/filtered_spy_data_10_yrs.csv')

class IntradaySpyEnv(SpyEnv):
    def __init__(self):
        self.date_str_format = '%Y-%m-%d %H:%M:%S%z'
        super().__init__()
    
    def _convert_date_str_to_datetime(self,datestr):
        if sys.version_info[1] <=5:
            datestr = re.sub(r'([-+]\d{2}):(\d{2})(?:(\d{2}))?$', r'\1\2\3', datestr)
        return datetime.strptime(datestr,self.date_str_format)
    
    def _load_data(self):
        data_file = os.path.join(os.path.dirname(__file__),'..','data/filtered_spy_2017_2019_all.csv')
        df = pd.read_csv(data_file)
        return df.columns.values, df.values


# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fingym.envs.spy_envs import DailySpyEnv, IntradaySpyEnv

import pytest

@pytest.fixture
def spy_daily_v0_env():
    return DailySpyEnv()

@pytest.fixture
def spy_intraday_minute_v0_env():
    return IntradaySpyEnv()

def test_make_spy_intraday_v0_env(spy_intraday_minute_v0_env):
    assert type(spy_intraday_minute_v0_env) == IntradaySpyEnv

def test_spy_intraday_minute_v0_init_values(spy_intraday_minute_v0_env):
    assert spy_intraday_minute_v0_env.cur_step == 0
    assert spy_intraday_minute_v0_env.n_step == len(spy_intraday_minute_v0_env.data)
    assert spy_intraday_minute_v0_env.initial_investment == 25000
    assert spy_intraday_minute_v0_env.cash_in_hand == spy_intraday_minute_v0_env.initial_investment
    assert spy_intraday_minute_v0_env.stock_owned == 0
    assert len(spy_intraday_minute_v0_env.headers) == 6
    assert spy_intraday_minute_v0_env.state_dim == 8
    assert spy_intraday_minute_v0_env._close_idx == 4
    assert spy_intraday_minute_v0_env._open_idx == 1

def test_spy_intraday_minute_v0_reset(spy_intraday_minute_v0_env):
    observation = spy_intraday_minute_v0_env.reset()
    assert spy_intraday_minute_v0_env.cur_step == 0
    assert spy_intraday_minute_v0_env.cash_in_hand == spy_intraday_minute_v0_env.initial_investment
    assert spy_intraday_minute_v0_env.stock_owned == 0
    assert len(observation) == spy_intraday_minute_v0_env.state_dim
    assert observation[0] == 0
    assert observation[1] == 25000
    assert observation[2] == 1483453800.0
    assert observation[3] == 225.04
    assert observation[4] == 225.12
    assert observation[5] == 224.93
    assert observation[6] == 224.95
    assert observation[7] == 774063

def test_make_spy_daily_v0(spy_daily_v0_env):
    assert type(spy_daily_v0_env) == DailySpyEnv

def test_spy_daily_v0_init_values(spy_daily_v0_env):
    assert spy_daily_v0_env.cur_step == 0
    assert spy_daily_v0_env.n_step == len(spy_daily_v0_env.data)
    assert spy_daily_v0_env.initial_investment == 25000
    assert spy_daily_v0_env.cash_in_hand == spy_daily_v0_env.initial_investment
    assert spy_daily_v0_env.stock_owned == 0
    assert len(spy_daily_v0_env.headers) == 6
    assert spy_daily_v0_env.state_dim == 8
    assert spy_daily_v0_env._close_idx == 4
    assert spy_daily_v0_env._open_idx == 1

def test_spy_daily_v0_reset(spy_daily_v0_env):
    observation = spy_daily_v0_env.reset()
    assert spy_daily_v0_env.cur_step == 0
    assert spy_daily_v0_env.cash_in_hand == spy_daily_v0_env.initial_investment
    assert spy_daily_v0_env.stock_owned == 0
    assert len(observation) == spy_daily_v0_env.state_dim
    assert observation[0] == 0
    assert observation[1] == 25000
    assert observation[2] == 1230019200.0
    assert observation[3] == 87.53
    assert observation[4] == 87.93
    assert observation[5] == 85.80
    assert observation[6] == 86.16
    assert observation[7] == 221772560

def test_spy_daily_v0_reset_after_action(spy_daily_v0_env):
    spy_daily_v0_env.reset()
    spy_daily_v0_env.step(spy_daily_v0_env.action_space.sample())
    observation = spy_daily_v0_env.reset()
    assert observation[0] == 0
    assert observation[1] == 25000
    assert observation[2] == 1230019200.0
    assert observation[3] == 87.53
    assert observation[4] == 87.93
    assert observation[5] == 85.80
    assert observation[6] == 86.16
    assert observation[7] == 221772560

# These are more SpyEnv testing common functionality. 

def test_spy_daily_v0_step(spy_daily_v0_env):
    spy_daily_v0_env.reset()
    obs, reward, done, info = spy_daily_v0_env.step(spy_daily_v0_env.action_space.sample())
    assert spy_daily_v0_env.cur_step == 1
    assert done == False
    assert info['cur_val'] == 25000
    assert reward == 0
    assert obs[0] == 0
    assert obs[1] == 25000
    assert obs[2] == 1230105600.0
    assert obs[3] == 86.45
    assert obs[4] == 86.87
    assert obs[5] == 86
    assert obs[6] == 86.66
    assert obs[7] == 62142416.0

def test_spy_daily_v0_buying_shares(spy_daily_v0_env):
    spy_daily_v0_env.reset()
    spy_daily_v0_env.step([1,10])
    assert spy_daily_v0_env.stock_owned == 10

def test_spy_daily_v0_selling_shares(spy_daily_v0_env):
    spy_daily_v0_env.reset()
    spy_daily_v0_env.step([1,10])
    spy_daily_v0_env.step([2,5])
    assert spy_daily_v0_env.stock_owned == 5

def test_cash_in_hand_reduced_when_buying_shares(spy_daily_v0_env):
    spy_daily_v0_env.reset()
    spy_daily_v0_env.step([1,10])
    assert spy_daily_v0_env.cash_in_hand == 24135.499999999993

def test_cash_in_hand_increased_when_selling_shares(spy_daily_v0_env):
    spy_daily_v0_env.reset()
    spy_daily_v0_env.step([1,10])
    spy_daily_v0_env.step([2,10])
    assert spy_daily_v0_env.cash_in_hand == 25007.899999999994

def test_cash_in_hand_increase_when_selling_shares_more_than_i_have(spy_daily_v0_env):
    spy_daily_v0_env.reset()
    spy_daily_v0_env.step([1,10])
    spy_daily_v0_env.step([2,11])
    assert spy_daily_v0_env.cash_in_hand == 25007.899999999994

def test_reward_from_step(spy_daily_v0_env):
    spy_daily_v0_env.reset()
    _, reward, _, _ = spy_daily_v0_env.step([1,10])
    assert reward == 2.099999999991269

def test_done(spy_daily_v0_env):
    spy_daily_v0_env.reset()
    done = False
    while not done:
        _, _, done, _ = spy_daily_v0_env.step(spy_daily_v0_env.action_space.sample())
    
    assert spy_daily_v0_env.cur_step == len(spy_daily_v0_env.data)-1

def test_done_doesnt_crash(spy_daily_v0_env):
    spy_daily_v0_env.reset()
    done = False
    while not done:
        _, _, done, _ = spy_daily_v0_env.step(spy_daily_v0_env.action_space.sample())
    
    spy_daily_v0_env.step(spy_daily_v0_env.action_space.sample())

    assert spy_daily_v0_env.cur_step == len(spy_daily_v0_env.data)-1


# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division
import argparse
from fingym import fingym

class BuyAndHoldAgent(object):
    def __init__(self, action_space):
        self.bought_yet = False
        self.action_space = action_space
    
    def act(self, observation, reward, done):
        if not self.bought_yet:
            
            cash_in_hand = observation[1]
            close_price = observation[6]
            num_shares_to_buy = cash_in_hand / close_price
            print('will buy {} shares'.format(num_shares_to_buy))
            self.bought_yet = True
            
            return [1,num_shares_to_buy]
        else:
            return [0,0]

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=None)
    parser.add_argument('env_id', nargs='?', default='SPY-Daily-v0', help='Select the environment to run')
    args = parser.parse_args()

    env = fingym.make(args.env_id)
    agent = BuyAndHoldAgent(env.action_space)

    reward = 0
    done = False

    cur_val = 0

    ob = env.reset()
    initial_value = ob[1]
    
    while True:
        action = agent.act(ob, reward, done)
        ob, reward, done, info = env.step(action)
        if done:
            cur_val = info['cur_val']
            break
    
    print('initial value: {}'.format(initial_value))
    print('final value: {}'.format(cur_val))


# import pandas as pd

# import os

# data = pd.read_csv(os.path.join(os.path.dirname(__file__),'amzn_full_weekly.csv'))
# data = data[['Date','Open','High','Low','Close','Volume']]
# data.set_index('Date',inplace=True)
# data.to_csv(os.path.join(os.path.dirname(__file__),'filtered_amzn_full_weekly.csv'))
# print(data.head)


# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fingym.envs.amzn_envs import AmznDailyEnv

import pytest

@pytest.fixture
def amzn_daily_v0_env():
    return AmznDailyEnv()

def test_make_amzn_daily_v0_env(amzn_daily_v0_env):
    assert type(amzn_daily_v0_env) == AmznDailyEnv

def test_amzn_daily_v0_file_location(amzn_daily_v0_env):
    assert 'filtered_amzn_data' in amzn_daily_v0_env._get_data_file()


# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque

import argparse
import os
import os.path

from fingym import fingym

import tensorflow as tf
from tensorflow.keras.models import model_from_json
from tensorflow.keras import Model, Sequential
from tensorflow.keras.layers import Dense, Embedding, Reshape
from tensorflow.keras.optimizers import Adam

import random
import numpy as np

class DQNAgent():

    def __init__(self, env_state_dim, time_frame, epsilon = 1, learning_rate=0.01, train = False):
        dirname = os.path.dirname(__file__)
        self.model_filepath = os.path.join(dirname,'weights.h5')
        self._state_size = env_state_dim
        self.trainMode = train

        # 0 - do nothing
        # 1 - buy w/ multiplier .33
        # 2 - buy w/ multiplier .5
        # 3 - buy w/ multiplier .66
        # 4 - sell w/ multiplier .33
        # 5 - sell w/ multiplier .5
        # 6 - sell w/ multiplier .66
        self._action_size = 7

        self.experience_replay = deque(maxlen=2000)

        self.gamma = 0.98
        if not self.trainMode:
            self.epsilon = 0
        else:
            self.epsilon = epsilon
        self.eps_decay = 0.995
        self.eps_min = 0.01

        self.max_shares_to_trade_at_once = 100

        # holds our last time_frame sequential state frames for prediction.
        self._time_frame = time_frame
        self.state_fifo = deque(maxlen=self._time_frame)

        # networks
        self.q_network = self._build_compile_model(learning_rate)
        self.target_network = self._build_compile_model(learning_rate)
        self._load_model_weights(self.q_network, self.model_filepath)
        self.align_target_model()

    def align_target_model(self):
        # reduce exploration rate as more
        # training happens
        if self.epsilon > self.eps_min:
            self.epsilon *= self.eps_decay
    
        print('epsilon: ', self.epsilon)

        self.target_network.set_weights(self.q_network.get_weights())
        self._save_model_weights(self.q_network, self.model_filepath)

    def _build_compile_model(self, learning_rate):
        '''
        Model taken from https://arxiv.org/pdf/1802.09477.pdf
        '''
        model = Sequential()

        # use a dense nn with inputs
        input_size = self._state_size * self._time_frame
        model.add(Dense(400, input_shape=(input_size,), activation='relu'))
        model.add(Dense(300, activation='relu'))
        #model.add(Dense(self._action_size, activation='tanh'))
        model.add(Dense(self._action_size, activation='linear'))
        model.compile(loss='mse', optimizer=Adam(learning_rate=learning_rate))
        print(model.summary())
        return model
    
    def store(self, state, action, reward, next_state, terminated):
        state = np.reshape(state, (self._state_size,self._time_frame))
        next_state = np.reshape(next_state, (self._state_size, self._time_frame))
        self.experience_replay.append((state, action, reward, next_state, terminated))
    
    def act(self, state):
        self.state_fifo.append(state)

        # do nothing for the first time frames until we can start the prediction
        if len(self.state_fifo) < self._time_frame:
            # Our environment takes a tuple for action https://entrpn.github.io/fingym/#spaces
            return np.zeros(2)
        
        # epsilon decays over time
        if np.random.rand() <= self.epsilon:
            return self._random_action()

        state = np.array(list(self.state_fifo))
        state = np.reshape(state,(self._state_size*self._time_frame,1))

        q_values = self.q_network.predict_on_batch(state.T)
        env_action = self._nn_action_to_env_action(np.argmax(q_values[0]))
        return env_action
    
    def retrain(self, batch_size):
        if not self.trainMode:
            return
        
        minibatch = random.sample(self.experience_replay, batch_size)

        for state, action, reward, next_state, terminated in minibatch:
            state = np.reshape(state,(self._state_size * self._time_frame,1))
            next_state = np.reshape(next_state, (self._state_size * self._time_frame, 1))
            target = np.array(self.q_network.predict_on_batch(state.T))

            if terminated[-1]:
                target[0][np.argmax(self._nn_action_to_env_action(action))] = reward[-1]
            else:
                t = np.array(self.target_network.predict_on_batch(next_state.T))
                target[0][np.argmax(self._env_action_to_nn_action(action))] = reward[-1] + self.gamma * np.amax(t)
            
            self.q_network.fit(state.T, target, epochs=1, verbose = 0)

    def _random_action(self):
        return self._nn_action_to_env_action(np.random.choice(self._action_size))
    
    def _nn_action_to_env_action(self,nn_action):
        env_action = [0,0]
        
        if nn_action == 0:
            env_action = [0,0]
        if nn_action == 1:
            env_action = [1, 0.33 * self.max_shares_to_trade_at_once]
        if nn_action == 2:
            env_action = [1, 0.5 * self.max_shares_to_trade_at_once]
        if nn_action == 3:
            env_action = [1, 0.66 * self.max_shares_to_trade_at_once]
        if nn_action == 4:
            env_action = [2, 0.33 * self.max_shares_to_trade_at_once]
        if nn_action == 5:
            env_action = [2, 0.5 * self.max_shares_to_trade_at_once]
        if nn_action == 6:
            env_action = [2, 0.66 * self.max_shares_to_trade_at_once]

        return env_action

    def _env_action_to_nn_action(self, action):
        nn_action = np.zeros(self._action_size)

        if action[0] == 0:
            nn_action[0] = 1

        if action[0] == 1:
            multiplier = action[1] / self.max_shares_to_trade_at_once
            if multiplier == 0.33:
                nn_action[1] = 1
            if multiplier == 0.5:
                nn_action[2] = 1
            if multiplier == 0.66:
                nn_action[3] = 1

        if action[0] == 2:
            multiplier = action[1] / self.max_shares_to_trade_at_once
            if multiplier == 0.33:
                nn_action[4] = 1
            if multiplier == 0.5:
                nn_action[5] = 1
            if multiplier == 0.66:
                nn_action[6] = 1

        return nn_action

    def _save_model_weights(self, model, filepath):
        model.save_weights(filepath)
    
    def _load_model_weights(self, model, filepath):
        if os.path.exists(filepath):
            print('loading existing weights')
            model.load_weights(filepath)

def reset_timeframe(time_frame, state_dim):
    s = np.zeros((time_frame,state_dim))
    r = np.zeros((time_frame))
    d = np.zeros((time_frame))
    ns = np.zeros((time_frame, state_dim))
    return s, r, ns, d 

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=None)
    parser.add_argument('env_id', nargs='?', default='SPY-Daily-v0', help='Select the environment to run')
    args = parser.parse_args()

    train = True
    if train:
        rang = 100
    else:
        rang = 1

    # collect the last 10 time frames (10 days for daily env) and use that to make a prediction for current action
    time_frame = 10
    time_frame_counter = 0

    # train on this batch size
    batch_size = 32

    env = fingym.make(args.env_id)
    # removing time element from state_dim since I'm creating a sequence via time_frame
    state_size = env.state_dim - 1
    print('state_size: ', state_size)
    agent = DQNAgent(state_size, time_frame, train = train)

    for i in range(rang):

        # init our env
        state = env.reset()
        # remove time element
        state = np.delete(state, 2)
        done = False

        # init our timeframe
        s_timeframe, r_timeframe, ns_timeframe, d_timeframe = reset_timeframe(time_frame, state_size)

        # alighn every training iterations
        align_every_itt = 15
        align_counter = 0

        while not done:
            action = agent.act(state)
            next_state, reward, done, info = env.step(action)
            print('action: ', action)
            print('reward: ', reward)

            # remove time element
            if len(state) > state_size:
                state = np.delete(state, 2)
            next_state = np.delete(next_state, 2)

            if time_frame_counter >= time_frame:
                agent.store(s_timeframe, action, r_timeframe, ns_timeframe, d_timeframe)
                s_timeframe[:-1] = s_timeframe[1:]
                r_timeframe[:-1] = r_timeframe[1:]
                ns_timeframe[:-1] = ns_timeframe[1:]
                d_timeframe[:-1] = d_timeframe[1:]
                time_frame_counter-=1
            
            s_timeframe[time_frame_counter] = state
            r_timeframe[time_frame_counter] = reward
            ns_timeframe[time_frame_counter] = next_state
            d_timeframe[time_frame_counter] = done
            time_frame_counter+=1

            if len(agent.experience_replay) > batch_size:
                print('retrain')
                agent.retrain(batch_size)

            if align_counter >= align_every_itt:
                print('align target model')
                agent.align_target_model()
                align_counter = 0
                print(info)
            
            state = next_state
            align_counter+=1



# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from fingym.spaces.space import BuyHoldSellSpace

@pytest.fixture
def buy_hold_sell_space():
    return BuyHoldSellSpace()

def test_buy_and_hold_sell_shape(buy_hold_sell_space):
    assert buy_hold_sell_space.shape == 2

def test_buy_hold_sell_sample_hold(buy_hold_sell_space):
    sample_action = buy_hold_sell_space.sample(2)
    assert sample_action[0] == 0
    assert sample_action[1] == 15

def test_buy_hold_sell_sample_buy(buy_hold_sell_space):
    sample_action = buy_hold_sell_space.sample(30)
    assert sample_action[0] == 1
    assert sample_action[1] == 37

def test_buy_hold_sell_sample_sell(buy_hold_sell_space):
    sample_action = buy_hold_sell_space.sample(4)
    assert sample_action[0] == 2
    assert sample_action[1] == 46


# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division
import argparse
from fingym import fingym

class RandomAgent(object):
    def __init__(self, action_space):
        self.action_space = action_space
    
    def act(self, observation, reward, done):
        return self.action_space.sample()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=None)
    parser.add_argument('env_id', nargs='?', default='SPY-Daily-v0', help='Select the environment to run')
    args = parser.parse_args()

    env = fingym.make(args.env_id)
    agent = RandomAgent(env.action_space)

    episode_count = 100
    reward = 0
    done = False

    final_vals = []

    initial_value = 0

    for i in range(episode_count):
        ob = env.reset()
        initial_value = ob[1]
        while True:
            action = agent.act(ob, reward, done)
            ob, reward, done, info = env.step(action)
            if done:
                final_vals.append(info['cur_val'])
                break
    
    max_value = max(final_vals)
    min_value = min(final_vals)
    avg_value = sum(final_vals)/len(final_vals)
    print('initial value: {}'.format(initial_value))
    print('min_value: {}, avg_value: {}, max_value: {}'.format(min_value,avg_value,max_value))


from fingym.examples.agents.multi_env_evo_agent import build_compile_model, EvoAgent

from fingym import fingym

import os

import numpy as np

import matplotlib.pyplot as plt

def test_agent(env, agent, model):

    state = env.reset()
    # Removed time element from state
    state = np.delete(state, 2)
    state_as_percentages = state
    done = False
    states_buy = []
    states_sell = []
    closes = []

    i = 0
    while not done:
        closes.append(state[5])
        action = agent.act(state_as_percentages, model)
        print(action)
        next_state, reward, done, info = env.step(action)
        if len(next_state) > agent.state_size:
            next_state = np.delete(next_state, 2)
        if action[0] == 1 and state[1] > next_state[2]:
            states_buy.append(i)
        if action[0] == 2 and state[0] > 0:
            states_sell.append(i)
        opn = (next_state[2] - state[2]) / next_state[2]
        high = (next_state[3] - state[3]) / next_state[3]
        low = (next_state[4] - state[4]) / next_state[4]
        close = (next_state[5] - state[5]) / next_state[5]
        volume = (next_state[6] - state[6]) / next_state[6]
        state_as_percentages = [next_state[0], next_state[1], opn, high, low, close, volume]
        state = next_state
        i+=1
    return closes, states_buy, states_sell, info['cur_val']

env = fingym.make('SPY-Daily-v0')
# removing time frame
state_size = env.state_dim - 1
time_frame = 30

dirname = os.path.dirname(__file__)
weights_file = os.path.join(dirname,'evo_weights_5.h5')
model = build_compile_model(state_size*time_frame,7)
model.load_weights(weights_file)

agent = EvoAgent(state_size, time_frame, model.get_weights())

closes, states_buy, states_sell, result = test_agent(env,agent,model)
print('result: {}'.format(str(result)))
plt.figure(figsize = (20, 10))
plt.plot(closes, label = 'true close', c = 'g')
plt.plot(closes, 'X', label = 'predict buy', markevery = states_buy, c = 'b')
plt.plot(closes, 'o', label = 'predict sell', markevery = states_sell, c = 'r')
plt.legend()
plt.show()


# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from fingym import fingym
from collections import deque

import matplotlib.pyplot as plt

import ray

import os

ray.init()

env = fingym.make('SPY-Daily-v0')

CONFIG = {
    'env_name': 'SPY-Daily-v0',
    # removing time frame, stocks owned and cash in hand
    'state_size': env.state_dim - 3,
    'max_shares_to_trade_at_once': 100,
    'time_frame': 30,
    'sigma': 0.1,
    'learning_rate': 0.03,
    'population_size': 400,
    'iterations': 50,
    'train': False,
    'eval': True,
    'log_actions': True
}

def get_state_as_change_percentage(state, next_state):
    open = (next_state[2] - state[2]) / next_state[2]
    high = (next_state[3] - state[3]) / next_state[3]
    low = (next_state[4] - state[4]) / next_state[4]
    close = (next_state[5] - state[5]) / next_state[5]
    volume = (next_state[6] - state[6]) / next_state[6]
    return [open, high, low, close, volume]

@ray.remote
def reward_function(weights):
    time_frame = CONFIG['time_frame']
    state_size = CONFIG['state_size']
    model = Model(time_frame * state_size, 500, 3)
    model.set_weights(weights)
    agent = Agent(model,state_size, time_frame)
    _,_,_,reward = run_agent(agent)
    print('reward: ',reward)
    return reward
    

def run_agent(agent):
    env = fingym.make(CONFIG['env_name'])
    log_actions = CONFIG['log_actions']
    state = env.reset()
    # Removed time element from state
    state = np.delete(state, 2)

    next_state, reward, done, info = env.step([0,0])
    if len(next_state) > agent.state_size:
        next_state = np.delete(next_state, 2)
    state_as_percentages = get_state_as_change_percentage(state,next_state)
    state = next_state

    done = False
    states_buy = []
    states_sell = []
    closes = []

    i = 0
    while not done:
        closes.append(state[5])
        action = agent.act(state_as_percentages)
        #if log_actions:
            #print('action: ',action)
            #print('state: ',state)
        next_state, reward, done, info = env.step(action)
        if len(next_state) > agent.state_size:
            next_state = np.delete(next_state, 2)
        if action[0] == 1 and action[1] > 0 and state[1] > state[2]:
            if log_actions:
                print('stocks owned: ',state[0])
                print('stocks to buy: ',action[1])
                print('stock price: ',state[2])
                print('cash in hand: ',state[1])
                print('total value: ',info['cur_val'])
            states_buy.append(i)
        if action[0] == 2 and action[1] > 0 and state[0] > 0:
            if log_actions:
                print('stocks owned: ',state[0])
                print('stocks to sell: ',action[1])
                print('stock price: ',state[2])
                print('cash in hand: ',state[1])
                print('total value: ',info['cur_val'])
            states_sell.append(i)
        state_as_percentages = get_state_as_change_percentage(state, next_state)
        state = next_state
        i+=1
    return closes, states_buy, states_sell, info['cur_val']

class Deep_Evolution_Strategy:
    def __init__(self, weights):
        self.weights = weights
        self.population_size = CONFIG['population_size']
        self.sigma = CONFIG['sigma']
        self.learning_rate = CONFIG['learning_rate']
    
    def _get_weight_from_population(self,weights, population):
        weights_population = []
        for index, i in enumerate(population):
            jittered = self.sigma * i
            weights_population.append(weights[index] + jittered)
        return weights_population
    
    def get_weights(self):
        return self.weights

    def train(self,epoch = 500, print_every=1):
        for i in range(epoch):
            population = []
            rewards = np.zeros(self.population_size)
            for k in range(self.population_size):
                x = []
                for w in self.weights:
                    x.append(np.random.randn(*w.shape))
                population.append(x)
            
            futures = [reward_function.remote(self._get_weight_from_population(self.weights,population[k])) for k in range(self.population_size)]

            rewards = ray.get(futures)
            
            rewards = (rewards - np.mean(rewards)) / np.std(rewards)
            for index, w in enumerate(self.weights):
                A = np.array([p[index] for p in population])
                self.weights[index] = (
                    w + self.learning_rate / (self.population_size * self.sigma) * np.dot(A.T, rewards).T
                )
            
            if (i + 1) % print_every == 0:
                print('iter: {}. standard reward: {}'.format(i+1,ray.get(reward_function.remote((self.weights)))))

class Agent:
    def __init__(self, model, state_size, time_frame):
        self.model = model
        self.time_frame = time_frame
        self.state_size = state_size
        self.state_fifo = deque(maxlen=self.time_frame)
        self.max_shares_to_trade_at_once = CONFIG['max_shares_to_trade_at_once']
        self.des = Deep_Evolution_Strategy(self.model.get_weights())
    
    def act(self,state):
        self.state_fifo.append(state)
        # do nothing for the first time frames until we can start the prediction
        if len(self.state_fifo) < self.time_frame:
            return np.zeros(2)
        
        state = np.array(list(self.state_fifo))
        state = np.reshape(state,(self.state_size*self.time_frame,1))
        #print(state)
        decision, buy = self.model.predict(state.T)
        # print('decision: ', decision)
        # print('buy: ', buy)

        return [np.argmax(decision[0]), min(self.max_shares_to_trade_at_once,max(int(buy[0]),0))]
    
    def fit(self, iterations, checkpoint):
        self.des.train(iterations, print_every = checkpoint)

class Model:
    def __init__(self, input_size, layer_size, output_size):
        self.weights = [
            np.random.randn(input_size, layer_size),
            np.random.randn(layer_size, output_size),
            np.random.randn(layer_size, 1),
            np.random.randn(1, layer_size)
        ]
    
    def predict(self, inputs):
        feed = np.dot(inputs, self.weights[0]) + self.weights[-1]
        decision = np.dot(feed, self.weights[1])
        buy = np.dot(feed, self.weights[2])
        return decision, buy
    
    def get_weights(self):
        return self.weights
    
    def set_weights(self, weights):
        self.weights = weights

if __name__ == '__main__':

    train = False
    eval = False

    time_frame = CONFIG['time_frame']
    state_size = CONFIG['state_size']
    
    model = Model(time_frame * state_size, 500, 3)

    dirname = os.path.dirname(__file__)
    weights_file = os.path.join(dirname,'deep_evo_weights.npy')

    if os.path.exists(weights_file):
        print('loading weights')
        weights = np.load(weights_file,allow_pickle=True)
        model.set_weights(weights)

    # np.save(weights_file,model.get_weights())
    agent = Agent(model,state_size, time_frame)
    if CONFIG['train']:
        agent.fit(iterations=CONFIG['iterations'], checkpoint=10)
        agent.model.set_weights(agent.des.get_weights())
        np.save(weights_file, agent.des.get_weights())
    
    if CONFIG['eval']:
        closes, states_buy, states_sell, result = run_agent(agent)
        print('result: {}'.format(str(result)))
        plt.figure(figsize = (20, 10))
        plt.plot(closes, label = 'true close', c = 'g')
        plt.plot(closes, 'X', label = 'predict buy', markevery = states_buy, c = 'b')
        plt.plot(closes, 'o', label = 'predict sell', markevery = states_sell, c = 'r')
        plt.legend()
        plt.show()


from setuptools import setup, find_packages, find_namespace_packages

import sys, os.path

# Don't import gym module here, since deps may not be installed
sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'fingym'))

setup(
  name = 'fingym',
  version = '0.3',
  license='apache-2.0',    
  description = 'A tool for developing reinforcement learning algorithms focused in stock prediction',
  author = 'Juan Acevedo',
  author_email = 'entrpn@gmail.com',
  url = 'https://github.com/entrpn/fingym/', 
  download_url = 'https://github.com/entrpn/fingym/archive/v0.3.tar.gz',
  keywords = ['stock-market', 'stock-price-prediction', 'python','reinforcement-learning-environments','reinforcement-learning','reinforcement-learning-agents','artificial-intelligence','q-learning','evolution-strategies'],
  packages=find_namespace_packages(),
  package_data={'fingym': ['data/*.csv']},
  install_requires=[
          'pandas',
          'numpy',
      ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers', 
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
  ],
)


# Copyright 2020 The fingym Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np
import itertools
from datetime import datetime, timedelta
import re
import sys
import os
from fingym.envs.spy_envs import DailySpyEnv, IntradaySpyEnv
from fingym.envs.tsla_envs import TslaDailyEnv
from fingym.envs.googl_envs import GooglDailyEnv
from fingym.envs.cgc_envs import CgcDailyEnv
from fingym.envs.cron_envs import CronDailyEnv
from fingym.envs.cgc_envs import CgcDailyEnv
from fingym.envs.ba_envs import BaDailyEnv
from fingym.envs.amzn_envs import AmznDailyEnv
from fingym.envs.amd_envs import AmdDailyEnv
from fingym.envs.abbv_envs import AbbvDailyEnv
from fingym.envs.aapl_envs import AaplDailyEnv

def make(envName):
    if envName == 'SPY-Daily-v0':
        return DailySpyEnv()
    if envName == 'SPY-Minute-v0':
        return IntradaySpyEnv()
    if envName == 'TSLA-Daily-v0':
        return TslaDailyEnv()
    if envName == 'GOOGL-Daily-v0':
        return GooglDailyEnv()
    if envName == 'CGC-Daily-v0':
        return CgcDailyEnv()
    if envName == 'CRON-Daily-v0':
        return CronDailyEnv()
    if envName == 'BA-Daily-v0':
        return BaDailyEnv()
    if envName == 'AMZN-Daily-v0':
        return AmznDailyEnv()
    if envName == 'AMD-Daily-v0':
        return AmdDailyEnv()
    if envName == 'ABBV-Daily-v0':
        return AbbvDailyEnv()
    if envName == 'AAPL-Daily-v0':
        return AaplDailyEnv()
    else:
        raise NotImplementedError